/*
 * earthapp.h
 *
 * $Id: earthapp.h,v 0.1 1998/12/10 20:08:01 thorsten Exp thorsten $
 *
 * This file is part of XGlobe. See README for details.
 *
 * Copyright (C) 1998 Thorsten Scheuermann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public Licenses as published by
 * the Free Software Foundation.
 *
 * $Log: earthapp.h,v $
 * Revision 0.1  1998/12/10 20:08:01  thorsten
 * initial revision
 *
 */

#ifndef _EARTHAPP_H
#define _EARTHAPP_H

#include <qapplication.h>
#include <qtimer.h>
#include <time.h>
#include "renderer.h"
#include "desktopwidget.h"
#include "markerlist.h"

class EarthApplication : public QApplication
{
  Q_OBJECT
    
public:
  EarthApplication(int &argc, char **argv);
  virtual ~EarthApplication();
  
  void init(void);

private:
  int readPosition(int i);
  void readDelay(int i);
  void readZoom(int i);
  void readPriority(int i);
  void readMarkerFile(int i);
  void readLabelPos(int i);
  void readAmbientLight(int i);
  void readMapFile(int i);
  void readNightMapFile(int i);  
  void readFov(int i);
  void printUsage(void);
  void printHelp(void);
  void randomPosition(void);
  void setPriority(int pri);
  
public slots:
  void recalc(void);

protected:
  enum PosType { FIXED, SUNREL, RANDOM };
 
  Renderer      *r;
  DesktopWidget *dwidget;
  QTimer        *timer;
  double        view_lat, view_long;
  double        zoom;
  int           delay;
  PosType       p_type;
  bool          builtin_markers;
  bool          show_markers;
  bool          show_label;
  int           label_x;
  int           label_y;
  double        ambient_light;
  double        fov;
  bool          with_nightmap;
  int           argc_map;
  int           argc_nightmap;
  time_t        time_to_render;
  MarkerList    marker_list;
  
private:
  bool firstTime;
  bool do_the_dump;
  bool use_kde;
};

#endif
