/*
 * renderer.h
 *
 * $Id: renderer.h,v 0.1 1998/12/09 18:10:25 thorsten Exp thorsten $
 *
 * This file is part of XGlobe. See README for details.
 *
 * Copyright (C) 1998 Thorsten Scheuermann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public Licenses as published by
 * the Free Software Foundation.
 *
 * $Log: renderer.h,v $
 * Revision 0.1  1998/12/09 18:10:25  thorsten
 * initial revision
 *
 */

#ifndef _RENDERER_H
#define _RENDERER_H

#include <qimage.h>
#include <qsize.h>
#include <qstring.h>
#include <qcolor.h>
#include <qpixmap.h>
#include <time.h>
#include "markerlist.h"

#ifndef PI
#define PI ((double)3.141592)
#endif

class Renderer
{
public:
  Renderer(const QSize &size, const char *mapfile = NULL);
  ~Renderer();
  int loadNightMap(const char *nmapfile = NULL);
  void renderFrame(void);
  void setViewPos(double lat, double lon);
  double getViewLat(void);
  double getViewLong(void);
  double getSunLat(void);
  double getSunLong(void);
  void setZoom(double z);
  double getZoom(void);
  void setTime(time_t t);
  time_t getTime(void);
  void setMarkerList(MarkerList *l);
  void showMarkers(bool show);
  void showLabel(bool show);
  void setLabelPos(int x, int y);
  void setAmbientLight(double ambient_light);
  void setFov(double fov);
  QImage *getImage(void);

protected:
  QImage *loadImage(const QString &name);
  
private:
  unsigned int getMapColorLinear(double longitude, double latitude,
                                 double angle);
  void         calcLightVector(void);
  void         calcDistance(void);
  void         drawMarkers(void);
  void         paintMarker(int x, int y, Location *l);
  void         drawLabel(void);
  static int   compareLocations(Location **l1, Location **l2);
  
protected:
  QImage     *map;
  QImage     *mapnight;
  QImage     *renderedImage;
  MarkerList *markerlist;
  bool       show_markers;
  bool       show_label;
  int        label_x;
  int        label_y;

private:
  QString mapFileName;
  QString mapNightFileName;
  bool    use_nightmap;

  QPixmap   *markerpixmap;
  
  // stuff used for rendering
  double		view_lat;
	double		view_long;
  double    sun_lat;
  double    sun_long;
  time_t    time_to_render;
	double		radius;
  double    proj_dist;          // distance to projection plane
  double    center_dist;        // distance to center of earth
  double    ambientLightLevel;  // how dark is the dark side?
	double		light_x,            // vector of sunlight with lengt 1
            light_y,
            light_z;
	double    fov;                // field of view
	double    zoom;
	int       radius_proj;        // radius of sphere on screen

};


#endif
